@@ -1,5 +1,5 @@
 --[[
-Copyright (c) 2011-2020, Vsevolod Stakhov <vsevolod@highsecure.ru>
+Copyright (c) 2022, Vsevolod Stakhov <vsevolod@rspamd.com>
 Copyright (c) 2013-2015, Andrew Lewis <nerf@judo.za.org>
 
 Licensed under the Apache License, Version 2.0 (the "License");
@@ -29,6 +29,7 @@ local selectors = require "lua_selectors"
 local bit = require 'bit'
 local lua_maps = require "lua_maps"
 local rbl_common = require "plugins/rbl"
+local rspamd_url = require "rspamd_url"
 
 -- This plugin implements various types of RBL checks
 -- Documentation can be found here:
@@ -43,23 +44,35 @@ local white_symbols = {}
 local black_symbols = {}
 local monitored_addresses = {}
 local known_selectors = {} -- map from selector string to selector id
+local url_flag_bits = rspamd_url.flags
 
 local function get_monitored(rbl)
+  local function is_random_monitored()
+    -- Explicit definition
+    if type(rbl.random_monitored) == 'boolean' then
+      return rbl.random_monitored
+    end
+
+    -- We check 127.0.0.1 for merely RBLs with `from` or `received` and only if
+    -- they don't have `no_ip` attribute at the same time
+    --
+    -- Convert to a boolean variable using the common idiom
+    return (not (rbl.from or rbl.received)
+        or rbl.no_ip)
+        and true or false
+  end
+
   local default_monitored = '1.0.0.127'
   local ret = {
     rcode = 'nxdomain',
     prefix = default_monitored,
-    random = false,
+    random = is_random_monitored(),
   }
 
   if rbl.monitored_address then
     ret.prefix = rbl.monitored_address
   end
 
-  if rbl.dkim or rbl.urls or rbl.emails then
-    ret.random = true
-  end
-
   lua_util.debugm(N, rspamd_config,
       'added monitored address: %s (%s random)',
       ret.prefix, ret.random)
@@ -123,28 +136,27 @@ local function ip_to_rbl(ip)
 end
 
 local function gen_check_rcvd_conditions(rbl, received_total)
-  local min_pos = tonumber(rbl['received_min_pos'])
-  local max_pos = tonumber(rbl['received_max_pos'])
-  local match_flags = rbl['received_flags']
-  local nmatch_flags = rbl['received_nflags']
+  local min_pos = tonumber(rbl.received_min_pos)
+  local max_pos = tonumber(rbl.received_max_pos)
+  local match_flags = rbl.received_flags
+  local nmatch_flags = rbl.received_nflags
+
   local function basic_received_check(rh)
-    if not (rh['real_ip'] and rh['real_ip']:is_valid()) then return false end
-    if ((rh['real_ip']:get_version() == 6 and rbl['ipv6']) or
-        (rh['real_ip']:get_version() == 4 and rbl['ipv4'])) and
-        ((rbl['exclude_private_ips'] and not rh['real_ip']:is_local()) or
-            not rbl['exclude_private_ips']) and ((rbl['exclude_local_ips'] and
-        not is_excluded_ip(rh['real_ip'])) or not rbl['exclude_local_ips']) then
+    if not (rh.real_ip and rh.real_ip:is_valid()) then return false end
+    if ((rh.real_ip:get_version() == 6 and rbl.ipv6) or
+        (rh.real_ip:get_version() == 4 and rbl.ipv4)) and
+        ((rbl.exclude_private_ips and not rh.real_ip:is_local()) or
+            not rbl.exclude_private_ips) and ((rbl.exclude_local_ips and
+        not is_excluded_ip(rh.real_ip)) or not rbl.exclude_local_ips) then
       return true
     else
       return false
     end
   end
-  if not (max_pos or min_pos or match_flags or nmatch_flags) then
-    return basic_received_check
-  end
-  return function(rh, pos)
-    if not basic_received_check() then return false end
-    local got_flags = rh['flags'] or E
+
+  local function positioned_received_check(rh, pos)
+    if not rh or not basic_received_check(rh) then return false end
+    local got_flags = rh.flags or E
     if min_pos then
       if min_pos < 0 then
         if min_pos == -1 then
@@ -152,7 +164,7 @@ local function gen_check_rcvd_conditions(rbl, received_total)
             return false
           end
         else
-          if pos <= (received_total - (min_pos*-1)) then
+          if pos <= (received_total - math.abs(min_pos)) then
             return false
           end
         end
@@ -162,7 +174,7 @@ local function gen_check_rcvd_conditions(rbl, received_total)
     end
     if max_pos then
       if max_pos < -1 then
-        if (received_total - (max_pos*-1)) >= pos then
+        if (received_total - math.abs(max_pos)) >= pos then
           return false
         end
       elseif max_pos > 0 then
@@ -187,6 +199,13 @@ local function gen_check_rcvd_conditions(rbl, received_total)
     end
     return true
   end
+
+
+  if not (max_pos or min_pos or match_flags or nmatch_flags) then
+    return basic_received_check
+  else
+    return positioned_received_check
+  end
 end
 
 local function rbl_dns_process(task, rbl, to_resolve, results, err, resolve_table_elt)
@@ -371,16 +390,16 @@ local function gen_rbl_callback(rule)
         end
       else
         local to_resolve
-        local orign = req
+        local origin = req
 
         if not resolve_ip then
-          orign = maybe_make_hash(req, rule)
+          origin = maybe_make_hash(req, rule)
           to_resolve = string.format('%s.%s',
-              orign,
+              origin,
               rule.rbl)
         else
           -- First, resolve origin stuff without hashing or anything
-          to_resolve = orign
+          to_resolve = origin
         end
 
         nreq = {
@@ -410,6 +429,14 @@ local function gen_rbl_callback(rule)
     return true
   end
 
+  local function check_required_symbols(task, _)
+    if rule.require_symbols then
+      return fun.all(function(sym) task:has_symbol(sym) end, rule.require_symbols)
+    end
+
+    return true
+  end
+
   local function check_user(task, _)
     if task:get_user() then
       return false
@@ -436,7 +463,8 @@ local function gen_rbl_callback(rule)
     local helo = task:get_helo()
 
     if not helo then
-      return false
+      -- Avoid pipeline breaking
+      return true
     end
 
     add_dns_request(task, helo, true, false, requests_table,
@@ -455,7 +483,8 @@ local function gen_rbl_callback(rule)
         -- We check merely mime from
         mime_from_domain = ((task:get_from('mime') or E)[1] or E).domain
         if mime_from_domain then
-          local mime_from_domain_tld = rspamd_util.get_tld(mime_from_domain)
+          local mime_from_domain_tld = rule.url_full_hostname and
+               mime_from_domain or rspamd_util.get_tld(mime_from_domain)
 
           if rule.url_compose_map then
             mime_from_domain = rule.url_compose_map:process_url(task, mime_from_domain_tld, mime_from_domain)
@@ -480,6 +509,8 @@ local function gen_rbl_callback(rule)
 
               if rule.url_compose_map then
                 domain_tld = rule.url_compose_map:process_url(task, domain_tld, domain)
+              elseif rule.url_full_hostname then
+                domain_tld = domain
               end
             end
 
@@ -492,6 +523,8 @@ local function gen_rbl_callback(rule)
               local domain_tld = rspamd_util.get_tld(domain)
               if rule.url_compose_map then
                 domain_tld = rule.url_compose_map:process_url(task, domain_tld, domain)
+              elseif rule.url_full_hostname then
+                domain_tld = domain
               end
               add_dns_request(task, domain_tld,
                   false, false, requests_table, 'dkim', whitelist)
@@ -539,9 +572,9 @@ local function gen_rbl_callback(rule)
     local urls = lua_util.extract_specific_urls(ex_params)
 
     for _,u in ipairs(urls) do
-      local flags = u:get_flags()
+      local flags = u:get_flags_num()
 
-      if flags.numeric then
+      if bit.band(flags, url_flag_bits.numeric) ~= 0 then
         -- For numeric urls we convert data to the ip address and
         -- reverse octets. See #3948 for details
         local to_resolve = u:get_host()
@@ -553,9 +586,10 @@ local function gen_rbl_callback(rule)
         add_dns_request(task, to_resolve, false,
             false, requests_table, 'url', whitelist)
       else
-        local url_tld = u:get_tld()
+        local url_hostname = u:get_host()
+        local url_tld = rule.url_full_hostname and url_hostname or u:get_tld()
         if rule.url_compose_map then
-          url_tld = rule.url_compose_map:process_url(task, url_tld, u:get_host())
+          url_tld = rule.url_compose_map:process_url(task, url_tld, url_hostname)
         end
         add_dns_request(task, url_tld, false,
             false, requests_table, 'url', whitelist)
@@ -603,7 +637,7 @@ local function gen_rbl_callback(rule)
   local function check_rdns(task, requests_table, whitelist)
     local hostname = task:get_hostname()
     if hostname == nil or hostname == 'unknown' then
-      return false
+      return true
     end
 
     add_dns_request(task, hostname, true, false,
@@ -635,6 +669,12 @@ local function gen_rbl_callback(rule)
     email_tbl.domain = email_tbl.domain:lower()
     email_tbl.user = email_tbl.user:lower()
 
+    if email_tbl.domain == '' or email_tbl.user == '' then
+      rspamd_logger.infox(task, "got an email with some empty parts: '%s@%s'; skip it in the checks",
+          email_tbl.user, email_tbl.domain)
+      return
+    end
+
     if rule.emails_domainonly then
       add_dns_request(task, email_tbl.domain, false, false, requests_table,
           what, whitelist)
@@ -681,7 +721,7 @@ local function gen_rbl_callback(rule)
 
     for _,email in ipairs(emails) do
       local domain
-      if rule.emails_domainonly then
+      if rule.emails_domainonly and not rule.url_full_hostname then
         if rule.url_compose_map then
           domain = rule.url_compose_map:process_url(task, email:get_tld(), email:get_host())
         else
@@ -692,8 +732,8 @@ local function gen_rbl_callback(rule)
       end
 
       local email_tbl = {
-        domain = domain,
-        user = email:get_user(),
+        domain = domain or '',
+        user = email:get_user() or '',
         addr = tostring(email),
       }
       check_email_table(task, email_tbl, requests_table, whitelist, 'email')
@@ -722,7 +762,8 @@ local function gen_rbl_callback(rule)
 
   -- Create function pipeline depending on rbl settings
   local pipeline = {
-    is_alive, -- generic for all
+    is_alive, -- check monitored status
+    check_required_symbols -- if we have require_symbols then check those symbols
   }
   local description = {
     'alive',
@@ -1000,6 +1041,7 @@ local function add_rbl(key, rbl, global_opts)
       id = rspamd_config:register_symbol{
         type = 'callback',
         callback = callback,
+        groups = {'rbl'},
         name = rbl.symbol .. '_CHECK',
         flags = table.concat(flags_tbl, ',')
       }
@@ -1026,6 +1068,7 @@ local function add_rbl(key, rbl, global_opts)
         type = 'callback',
         callback = callback,
         name = rbl.symbol,
+        groups = {'rbl'},
         group = 'rbl',
         score = 0,
         flags = table.concat(flags_tbl, ',')
@@ -1160,7 +1203,20 @@ if(opts['local_exclude_ip_map'] ~= nil) then
     'RBL exclusions map')
 end
 
-for key,rbl in pairs(opts.rbls ) do
+-- TODO: this code should be universal for all modules that use selectors to allow
+-- maps usage from selectors registered for a specific module
+if type(opts.attached_maps) == 'table' then
+  opts.attached_maps_processed = {}
+  for i,map in ipairs(opts.attached_maps) do
+    -- Store maps in the configuration table to keep lifetime track
+    opts.attached_maps_processed[i] = lua_maps.map_add_from_ucl(map)
+    if opts.attached_maps_processed[i] == nil then
+      rspamd_logger.warnx(rspamd_config, "cannot parse attached map: %s", map)
+    end
+  end
+end
+
+for key,rbl in pairs(opts.rbls) do
   if type(rbl) ~= 'table' or rbl.disabled == true or rbl.enabled == false then
     rspamd_logger.infox(rspamd_config, 'disable rbl "%s"', key)
   else
@@ -1243,14 +1299,18 @@ rspamd_config:register_symbol{
   type = 'callback',
   callback = rbl_callback_white,
   name = 'RBL_CALLBACK_WHITE',
-  flags = 'nice,empty,no_squeeze'
+  flags = 'nice,empty,no_squeeze',
+  groups = {'rbl'},
+  augmentations = {string.format("timeout=%f", rspamd_config:get_dns_timeout() or 0.0)},
 }
 
 rspamd_config:register_symbol{
   type = 'callback',
   callback = rbl_callback_fin,
   name = 'RBL_CALLBACK',
-  flags = 'empty,no_squeeze'
+  flags = 'empty,no_squeeze',
+  groups = {'rbl'},
+  augmentations = {string.format("timeout=%f", rspamd_config:get_dns_timeout() or 0.0)},
 }
 
 for _, w in ipairs(white_symbols) do
